@@ -1,6 +1,3 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from datetime import datetime
 from sys import exit
 
@@ -28,17 +25,24 @@ from ..utils.ui import io
 
 def bw_apply(repo, args):
     errors = []
-    target_nodes = get_target_nodes(repo, args['target'], adhoc_nodes=args['adhoc_nodes'])
-    pending_nodes = target_nodes[:]
+    target_nodes = get_target_nodes(repo, args['targets'])
+    pending_nodes = target_nodes.copy()
 
     io.progress_set_total(count_items(pending_nodes))
 
-    repo.hooks.apply_start(
-        repo,
-        args['target'],
-        target_nodes,
-        interactive=args['interactive'],
-    )
+    try:
+        repo.hooks.apply_start(
+            repo,
+            args['targets'],
+            target_nodes,
+            interactive=args['interactive'],
+        )
+    except GracefulApplyException as exc:
+        io.stderr(_("{x} apply aborted by hook ({reason})").format(
+            reason=str(exc) or _("no reason given"),
+            x=red("!!!"),
+        ))
+        exit(1)
 
     start_time = datetime.now()
     results = []
@@ -98,22 +102,23 @@ def bw_apply(repo, args):
     worker_pool.run()
 
     total_duration = datetime.now() - start_time
+    totals = stats(results)
 
     if args['summary'] and results:
-        stats_summary(results, total_duration)
+        stats_summary(results, totals, total_duration)
     error_summary(errors)
 
     repo.hooks.apply_end(
         repo,
-        args['target'],
+        args['targets'],
         target_nodes,
         duration=total_duration,
     )
 
-    exit(1 if errors else 0)
+    exit(1 if errors or totals['failed'] else 0)
 
 
-def stats_summary(results, total_duration):
+def stats(results):
     totals = {
         'items': 0,
         'correct': 0,
@@ -121,7 +126,14 @@ def stats_summary(results, total_duration):
         'skipped': 0,
         'failed': 0,
     }
+    for result in results:
+        totals['items'] += result.total
+        for metric in ('correct', 'fixed', 'skipped', 'failed'):
+            totals[metric] += getattr(result, metric)
+    return totals
 
+
+def stats_summary(results, totals, total_duration):
     rows = [[
         bold(_("node")),
         _("items"),
@@ -133,9 +145,6 @@ def stats_summary(results, total_duration):
     ], ROW_SEPARATOR]
 
     for result in results:
-        totals['items'] += result.total
-        for metric in ('correct', 'fixed', 'skipped', 'failed'):
-            totals[metric] += getattr(result, metric)
         rows.append([
             result.node_name,
             str(result.total),
