@@ -1,6 +1,3 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from os import makedirs
 from os.path import dirname, exists, join
 from sys import exit
@@ -27,7 +24,7 @@ def write_preview(file_item, base_path):
 
 
 def bw_items(repo, args):
-    node = get_node(repo, args['node'], adhoc_nodes=args['adhoc_nodes'])
+    node = get_node(repo, args['node'])
     if args['preview'] and not args['item']:
         io.stderr(_("{x} no ITEM given for preview").format(x=red("!!!")))
         exit(1)
