@@ -1,6 +1,3 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from .exceptions import BundleError, ItemDependencyError, NoSuchItem
 from .items import Item
 from .items.actions import Action
@@ -8,8 +5,9 @@ from .utils.text import bold, mark_for_translation as _
 from .utils.ui import io
 
 
-class DummyItem(object):
+class DummyItem:
     bundle = None
+    cascade_skip = True
     triggered = False
 
     def __init__(self, *args, **kwargs):
@@ -141,18 +139,24 @@ def _flatten_deps_for_item(item, items):
     """
     item._flattened_deps = set(item._deps)
 
-    for dep in item._deps:
+    for dep in item._deps.copy():
         try:
             dep_item = items[dep]
         except KeyError:
-            raise ItemDependencyError(_(
-                "'{item}' in bundle '{bundle}' has a dependency (needs) "
-                "on '{dep}', which doesn't exist"
-            ).format(
-                item=item.id,
-                bundle=item.bundle.name,
-                dep=dep,
-            ))
+            if dep.startswith("tag:"):
+                # sometimes it is useful to be able to depend on a tag
+                # without having to make sure it actually exists
+                item._deps.remove(dep)
+                continue
+            else:
+                raise ItemDependencyError(_(
+                    "'{item}' in bundle '{bundle}' has a dependency (needs) "
+                    "on '{dep}', which doesn't exist"
+                ).format(
+                    item=item.id,
+                    bundle=item.bundle.name,
+                    dep=dep,
+                ))
         # Don't recurse if we have already resolved nested dependencies
         # for this item. Also serves as a guard against infinite
         # recursion when there are loops.
@@ -665,9 +669,12 @@ def remove_item_dependents(items, dep_item, skipped=False):
 
     all_recursively_removed_items = []
     for removed_item in removed_items:
-        items, recursively_removed_items = \
-            remove_item_dependents(items, removed_item, skipped=skipped)
-        all_recursively_removed_items += recursively_removed_items
+        if removed_item.cascade_skip:
+            items, recursively_removed_items = \
+                remove_item_dependents(items, removed_item, skipped=skipped)
+            all_recursively_removed_items += recursively_removed_items
+        else:
+            items = remove_dep_from_items(items, removed_item.id)
 
     return (items, removed_items + all_recursively_removed_items)
 
