@@ -1,16 +1,11 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from abc import ABCMeta, abstractmethod
 
 from bundlewrap.exceptions import BundleError
 from bundlewrap.items import Item
 from bundlewrap.utils.text import mark_for_translation as _
-from six import add_metaclass
 
 
-@add_metaclass(ABCMeta)
-class Pkg(Item):
+class Pkg(Item, metaclass=ABCMeta):
     """
     A generic package.
     """
