@@ -1,6 +1,3 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from copy import copy
 from hashlib import sha1
 from json import dumps, JSONEncoder
@@ -11,27 +8,69 @@ from .utils.dicts import ATOMIC_TYPES, map_dict_keys, merge_dict, value_at_key_p
 from .utils.text import force_text, mark_for_translation as _
 
 
-try:
-    text_type = unicode
-    byte_type = str
-except NameError:
-    text_type = str
-    byte_type = bytes
-
-METADATA_TYPES = (
+METADATA_TYPES = (  # only meant for natively atomic types
     bool,
-    byte_type,
+    bytes,
     Fault,
     int,
-    text_type,
+    str,
     type(None),
 )
 
-# constants returned as options by metadata processors
-DONE = 1
-RUN_ME_AGAIN = 2
-DEFAULTS = 3
-OVERWRITE = 4
+
+class DoNotRunAgain(Exception):
+    """
+    Raised from metadata reactors to indicate they can be disregarded.
+    """
+    pass
+
+
+def deepcopy_metadata(obj):
+    """
+    Our own version of deepcopy.copy that doesn't pickle.
+    """
+    if isinstance(obj, METADATA_TYPES):
+        return obj
+    elif isinstance(obj, dict):
+        if isinstance(obj, ATOMIC_TYPES[dict]):
+            new_obj = atomic({})
+        else:
+            new_obj = {}
+        for key, value in obj.items():
+            new_key = copy(key)
+            new_obj[new_key] = deepcopy_metadata(value)
+    elif isinstance(obj, (list, tuple)):
+        if isinstance(obj, (ATOMIC_TYPES[list], ATOMIC_TYPES[tuple])):
+            new_obj = atomic([])
+        else:
+            new_obj = []
+        for member in obj:
+            new_obj.append(deepcopy_metadata(member))
+    elif isinstance(obj, set):
+        if isinstance(obj, ATOMIC_TYPES[set]):
+            new_obj = atomic(set())
+        else:
+            new_obj = set()
+        for member in obj:
+            new_obj.add(deepcopy_metadata(member))
+    else:
+        assert False  # there should be no other types
+    return new_obj
+
+
+def validate_metadata(metadata, _top_level=True):
+    if _top_level and not isinstance(metadata, dict):
+        raise TypeError(_("metadata must be a dict"))
+    if isinstance(metadata, dict):
+        for key, value in metadata.items():
+            if not isinstance(key, str):
+                raise TypeError(_("metadata keys must be str, not: {}").format(repr(key)))
+            validate_metadata(value, _top_level=False)
+    elif isinstance(metadata, (tuple, list, set)):
+        for value in metadata:
+            validate_metadata(value, _top_level=False)
+    elif not isinstance(metadata, METADATA_TYPES):
+        raise TypeError(_("illegal metadata value type: {}").format(repr(metadata)))
 
 
 def atomic(obj):
@@ -49,121 +88,68 @@ def atomic(obj):
         return cls(obj)
 
 
-def blame_changed_paths(old_dict, new_dict, blame_dict, blame_name, defaults=False):
-    def is_mergeable(value1, value2):
-        if isinstance(value1, (list, set, tuple)) and isinstance(value2, (list, set, tuple)):
-            return True
-        elif isinstance(value1, dict) and isinstance(value2, dict):
-            return True
-        return False
-
-    new_paths = map_dict_keys(new_dict)
-
-    # clean up removed paths from blame_dict
-    for path in list(blame_dict.keys()):
-        if path not in new_paths:
-            del blame_dict[path]
-
-    for path in new_paths:
-        new_value = value_at_key_path(new_dict, path)
-        try:
-            old_value = value_at_key_path(old_dict, path)
-        except KeyError:
-            blame_dict[path] = (blame_name,)
-        else:
-            if old_value != new_value:
-                if defaults or is_mergeable(old_value, new_value):
-                    blame_dict[path] += (blame_name,)
-                else:
-                    blame_dict[path] = (blame_name,)
-    return blame_dict
+def check_for_metadata_conflicts(node):
+    check_for_metadata_conflicts_between_groups(node)
+    check_for_metadata_conflicts_between_defaults_and_reactors(node)
 
 
-def check_metadata_keys(node):
-    try:
-        basestring
-    except NameError:  # Python 2
-        basestring = str
-    for path in map_dict_keys(node.metadata):
-        value = path[-1]
-        if not isinstance(value, basestring):
-            raise TypeError(_("metadata key for {node} at path '{path}' is not a string").format(
-                node=node.name,
-                path="'->'".join(path[:-1]),
-            ))
-
-
-def check_metadata_processor_result(input_metadata, result, node_name, metadata_processor_name):
+def check_for_metadata_conflicts_between_defaults_and_reactors(node):
     """
-    Validates the return value of a metadata processor and splits it
-    into metadata and options.
+    Finds conflicting metadata keys in bundle defaults and reactors.
+
+    Dicts can be merged with dicts, sets can be merged with sets, but
+    any other combination is a conflict.
     """
-    if not isinstance(result, tuple) or not len(result) >= 2:
-        raise ValueError(_(
-            "metadata processor {metaproc} for node {node} did not return "
-            "a tuple of length 2 or greater"
-        ).format(
-            metaproc=metadata_processor_name,
-            node=node_name,
-        ))
-    result_dict, options = result[0], result[1:]
-    if not isinstance(result_dict, dict):
-        raise ValueError(_(
-            "metadata processor {metaproc} for node {node} did not return "
-            "a dict as the first element"
-        ).format(
-            metaproc=metadata_processor_name,
-            node=node_name,
-        ))
-    if (
-        (DEFAULTS in options or OVERWRITE in options) and
-        id(input_metadata) == id(result_dict)
-    ):
-        raise ValueError(_(
-            "metadata processor {metaproc} for node {node} returned original "
-            "metadata dict plus DEFAULTS or OVERWRITE"
-        ).format(
-            metaproc=metadata_processor_name,
-            node=node_name,
-        ))
-    for option in options:
-        if option not in (DEFAULTS, DONE, RUN_ME_AGAIN, OVERWRITE):
-            raise ValueError(_(
-                "metadata processor {metaproc} for node {node} returned an "
-                "invalid option: {opt}"
-            ).format(
-                metaproc=metadata_processor_name,
-                node=node_name,
-                opt=repr(option),
-            ))
-    if DONE in options and RUN_ME_AGAIN in options:
-        raise ValueError(_(
-            "metadata processor {metaproc} for node {node} cannot return both "
-            "DONE and RUN_ME_AGAIN"
-        ).format(
-            metaproc=metadata_processor_name,
-            node=node_name,
-        ))
-    if DONE not in options and RUN_ME_AGAIN not in options:
-        raise ValueError(_(
-            "metadata processor {metaproc} for node {node} must return either "
-            "DONE or RUN_ME_AGAIN"
-        ).format(
-            metaproc=metadata_processor_name,
-            node=node_name,
-        ))
-    if DEFAULTS in options and OVERWRITE in options:
-        raise ValueError(_(
-            "metadata processor {metaproc} for node {node} cannot return both "
-            "DEFAULTS and OVERWRITE"
-        ).format(
-            metaproc=metadata_processor_name,
-            node=node_name,
-        ))
-    return result_dict, options
+    TYPE_DICT = 1
+    TYPE_SET = 2
+    TYPE_OTHER = 3
+
+    def paths_with_values_and_types(d):
+        for path in map_dict_keys(d):
+            value = value_at_key_path(d, path)
+            if isinstance(value, dict):
+                yield path, value, TYPE_DICT
+            elif isinstance(value, set):
+                yield path, value, TYPE_SET
+            else:
+                yield path, value, TYPE_OTHER
+
+    for prefix in ("metadata_defaults:", "metadata_reactor:"):
+        paths = {}
+        for partition in node._metadata_stack._partitions:
+            for identifier, layer in partition.items():
+                if identifier.startswith(prefix):
+                    for path, value, current_type in paths_with_values_and_types(layer):
+                        try:
+                            prev_type, prev_identifier, prev_value = paths[path]
+                        except KeyError:
+                            paths[path] = current_type, identifier, value
+                        else:
+                            if (
+                                prev_type == TYPE_DICT
+                                and current_type == TYPE_DICT
+                            ):
+                                pass
+                            elif (
+                                prev_type == TYPE_SET
+                                and current_type == TYPE_SET
+                            ):
+                                pass
+                            elif value != prev_value:
+                                raise ValueError(_(
+                                    "{node}: {a} and {b} are clashing over this key path: {path} "
+                                    "(\"{val_a}\" vs. \"{val_b}\")"
+                                ).format(
+                                    a=identifier,
+                                    b=prev_identifier,
+                                    node=node.name,
+                                    path="/".join(path),
+                                    val_a=value,
+                                    val_b=prev_value,
+                                ))
 
 
-def check_for_unsolvable_metadata_key_conflicts(node):
+def check_for_metadata_conflicts_between_groups(node):
     """
     Finds metadata keys defined by two groups that are not part of a
     shared subgroup hierarchy.
@@ -233,7 +219,7 @@ def check_for_unsolvable_metadata_key_conflicts(node):
     for chain in chains:
         metadata = {}
         for group in chain:
-            metadata = merge_dict(metadata, group.metadata)
+            metadata = merge_dict(metadata, group._attributes.get('metadata', {}))
         chain_metadata.append(metadata)
 
     # create a "key path map" for each chain's metadata
@@ -270,49 +256,17 @@ def check_for_unsolvable_metadata_key_conflicts(node):
                         )
 
 
-def deepcopy_metadata(obj):
-    """
-    Our own version of deepcopy.copy that doesn't pickle and ensures
-    a limited range of types is used in metadata.
-    """
-    if isinstance(obj, METADATA_TYPES):
-        return obj
-    elif isinstance(obj, dict):
-        if isinstance(obj, ATOMIC_TYPES[dict]):
-            new_obj = atomic({})
-        else:
-            new_obj = {}
-        for key, value in obj.items():
-            if not isinstance(key, METADATA_TYPES):
-                raise ValueError(_("illegal metadata key type: {}").format(repr(key)))
-            new_key = copy(key)
-            new_obj[new_key] = deepcopy_metadata(value)
-    elif isinstance(obj, (list, tuple)):
-        if isinstance(obj, (ATOMIC_TYPES[list], ATOMIC_TYPES[tuple])):
-            new_obj = atomic([])
-        else:
-            new_obj = []
-        for member in obj:
-            new_obj.append(deepcopy_metadata(member))
-    elif isinstance(obj, set):
-        if isinstance(obj, ATOMIC_TYPES[set]):
-            new_obj = atomic(set())
-        else:
-            new_obj = set()
-        for member in obj:
-            new_obj.add(deepcopy_metadata(member))
-    else:
-        raise ValueError(_("illegal metadata value type: {}").format(repr(obj)))
-    return new_obj
-
-
 def find_groups_causing_metadata_conflict(node_name, chain1, chain2, keypath):
     """
     Given two chains (lists of groups), find one group in each chain
     that has conflicting metadata with the other for the given key path.
     """
-    chain1_metadata = [list(map_dict_keys(group.metadata)) for group in chain1]
-    chain2_metadata = [list(map_dict_keys(group.metadata)) for group in chain2]
+    chain1_metadata = [
+        list(map_dict_keys(group._attributes.get('metadata', {}))) for group in chain1
+    ]
+    chain2_metadata = [
+        list(map_dict_keys(group._attributes.get('metadata', {}))) for group in chain2
+    ]
 
     bad_keypath = None
 
@@ -358,13 +312,17 @@ class MetadataJSONEncoder(JSONEncoder):
             raise ValueError(_("illegal metadata value type: {}").format(repr(obj)))
 
 
+def metadata_to_json(metadata, sort_keys=True):
+    return dumps(
+        metadata,
+        cls=MetadataJSONEncoder,
+        indent=4,
+        sort_keys=sort_keys,
+    )
+
+
 def hash_metadata(sdict):
     """
     Returns a canonical SHA1 hash to describe this dict.
     """
-    return sha1(dumps(
-        sdict,
-        cls=MetadataJSONEncoder,
-        indent=None,
-        sort_keys=True,
-    ).encode('utf-8')).hexdigest()
+    return sha1(metadata_to_json(sdict).encode('utf-8')).hexdigest()
