@@ -1,11 +1,11 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from datetime import datetime, timedelta
 from hashlib import md5
-from os import environ
+from os import environ, mkdir
+from os.path import exists, join
 from threading import Lock
 
+from tomlkit import dumps as toml_dump, parse as toml_parse
+
 from . import operations
 from .bundle import Bundle
 from .concurrency import WorkerPool
@@ -14,7 +14,6 @@ from .deps import (
     find_item,
 )
 from .exceptions import (
-    DontCache,
     GracefulApplyException,
     ItemDependencyLoop,
     NodeLockedException,
@@ -23,13 +22,20 @@ from .exceptions import (
     RepositoryError,
     SkipNode,
 )
-from .group import GROUP_ATTR_DEFAULTS
+from .group import GROUP_ATTR_DEFAULTS, GROUP_ATTR_TYPES
 from .itemqueue import ItemQueue
 from .items import Item
 from .lock import NodeLock
 from .metadata import hash_metadata
-from .utils import cached_property, names
-from .utils.dicts import hash_statedict
+from .utils import cached_property, error_context, get_file_contents, names, NO_DEFAULT
+from .utils.dicts import (
+    dict_to_toml,
+    hash_statedict,
+    set_key_at_path,
+    validate_dict,
+    value_at_key_path,
+    COLLECTION_OF_STRINGS,
+)
 from .utils.text import (
     blue,
     bold,
@@ -39,13 +45,19 @@ from .utils.text import (
     green,
     mark_for_translation as _,
     red,
+    toml_clean,
     validate_name,
     yellow,
 )
 from .utils.ui import io
 
 
-class ApplyResult(object):
+NODE_ATTR_TYPES = GROUP_ATTR_TYPES.copy()
+NODE_ATTR_TYPES['groups'] = COLLECTION_OF_STRINGS
+NODE_ATTR_TYPES['hostname'] = str
+
+
+class ApplyResult:
     """
     Holds information about an apply run for a node.
     """
@@ -122,6 +134,10 @@ def handle_apply_result(node, item, status_code, interactive, details=None):
     if formatted_result is not None:
         if status_code == Item.STATUS_FAILED:
             io.stderr(formatted_result)
+            if item._command_results:
+                io.stderr(format_item_command_results(item._command_results))
+                # free up memory
+                del item._command_results
         else:
             io.stdout(formatted_result)
 
@@ -270,6 +286,38 @@ def _flatten_group_hierarchy(groups):
     return order
 
 
+def format_item_command_results(results):
+    output = ""
+
+    for i in range(len(results)):
+        stdout = results[i]['result'].stdout_text.strip()
+        stderr = results[i]['result'].stderr_text.strip()
+
+        # show command
+        output += "\n{b}".format(b=red('│'))
+        output += "\n{b} {command} (return code: {code}{no_output})".format(
+            b=red('├─'),
+            command=bold(results[i]['command']),
+            code=bold(results[i]['result'].return_code),
+            no_output='' if stdout or stderr else '; no output'
+        )
+
+        # show output
+        lines = []
+        if stdout or stderr:
+            output += "\n{b}".format(b=red("│ "))
+            if stdout:
+                lines += stdout.strip().split('\n')
+            if stderr:
+                lines += stderr.strip().split('\n')
+
+        for k in range(len(lines)):
+            output += "\n{b} {line}".format(b=red("│ "), line=lines[k])
+
+    output += red("\n╵ ")
+    return output.lstrip('\n')
+
+
 def format_item_result(result, node, bundle, item, interactive=False, details=None):
     if details is True:
         details_text = "({})".format(_("create"))
@@ -318,7 +366,7 @@ def format_item_result(result, node, bundle, item, interactive=False, details=No
         )
 
 
-class Node(object):
+class Node:
     OS_FAMILY_BSD = (
         'freebsd',
         'macos',
@@ -358,16 +406,14 @@ class Node(object):
         if not validate_name(name):
             raise RepositoryError(_("'{}' is not a valid node name").format(name))
 
+        with error_context(node_name=name):
+            validate_dict(attributes, NODE_ATTR_TYPES)
+
         self._add_host_keys = environ.get('BW_ADD_HOST_KEYS', False) == "1"
-        self._bundles = attributes.get('bundles', [])
-        self._compiling_metadata = Lock()
-        self._dynamic_group_lock = Lock()
-        self._dynamic_groups_resolved = False  # None means we're currently doing it
-        self._metadata_so_far = {}
-        self._node_metadata = attributes.get('metadata', {})
+        self._attributes = attributes
         self._ssh_conn_established = False
         self._ssh_first_conn_lock = Lock()
-        self._template_node_name = attributes.get('template_node')
+        self.file_path = attributes.get('file_path')
         self.hostname = attributes.get('hostname', name)
         self.name = name
 
@@ -382,31 +428,23 @@ class Node(object):
 
     @cached_property
     def bundles(self):
-        if self._dynamic_group_lock.acquire(False):
-            self._dynamic_group_lock.release()
-        else:
-            raise RepositoryError(_(
-                "node bundles cannot be queried with members_add/remove"
-            ))
         with io.job(_("{node}  loading bundles").format(node=bold(self.name))):
-            added_bundles = []
-            found_bundles = []
-            for group in self.groups:
-                for bundle_name in group.bundle_names:
-                    found_bundles.append(bundle_name)
+            bundle_names = set(self._attributes.get('bundles', set()))
 
-            for bundle_name in found_bundles + list(self._bundles):
-                if bundle_name not in added_bundles:
-                    added_bundles.append(bundle_name)
-                    try:
-                        yield Bundle(self, bundle_name)
-                    except NoSuchBundle:
-                        raise NoSuchBundle(_(
-                            "Node '{node}' wants bundle '{bundle}', but it doesn't exist."
-                        ).format(
-                            bundle=bundle_name,
-                            node=self.name,
-                        ))
+            for group in self.groups:
+                for bundle_name in set(group._attributes.get('bundles', set())):
+                    bundle_names.add(bundle_name)
+
+            for bundle_name in bundle_names:
+                try:
+                    yield Bundle(self, bundle_name)
+                except NoSuchBundle:
+                    raise NoSuchBundle(_(
+                        "Node '{node}' wants bundle '{bundle}', but it doesn't exist."
+                    ).format(
+                        bundle=bundle_name,
+                        node=self.name,
+                    ))
 
     @cached_property
     def cdict(self):
@@ -423,7 +461,7 @@ class Node(object):
         True if this node should be skipped based on the given selector
         string (e.g. "node:foo,group:bar").
         """
-        components = [c.strip() for c in autoskip_selector.split(",")]
+        components = [c.strip() for c in autoskip_selector]
         if "node:{}".format(self.name) in components:
             return True
         for group in self.groups:
@@ -437,30 +475,23 @@ class Node(object):
     @cached_property
     @io.job_wrapper(_("{}  determining groups").format(bold("{0.name}")))
     def groups(self):
-        _groups = set(self.repo._static_groups_for_node(self))
-        # lock to avoid infinite recursion when .members_add/remove
-        # use stuff like node.in_group() that in turn calls this function
-        if self._dynamic_group_lock.acquire(False):
-            cache_result = True
-            self._dynamic_groups_resolved = None
-            # first we remove ourselves from all static groups whose
-            # .members_remove matches us
-            for group in list(_groups):
-                if group.members_remove is not None and group.members_remove(self):
-                    try:
-                        _groups.remove(group)
-                    except KeyError:
-                        pass
-            # now add all groups whose .members_add (but not .members_remove)
-            # matches us
-            _groups = _groups.union(self._groups_dynamic)
-            self._dynamic_groups_resolved = True
-            self._dynamic_group_lock.release()
-        else:
-            cache_result = False
+        _groups = set()
+
+        for group_name in set(self._attributes.get('groups', set())):
+            with error_context(node=self.name):
+                _groups.add(self.repo.get_group(group_name))
+
+        for group in self.repo.groups:
+            if group in _groups:
+                # we're already in this group, no need to check it again
+                continue
+            if self in group._nodes_from_members:
+                _groups.add(group)
+                continue
+            for pattern in group._member_patterns:
+                if pattern.search(self.name) is not None:
+                    _groups.add(group)
 
-        # we have to add parent groups at the very end, since we might
-        # have added or removed subgroups thru .members_add/remove
         while True:
             # Since we're only looking at *immediate* parent groups,
             # we have to keep doing this until we stop adding parent
@@ -468,40 +499,11 @@ class Node(object):
             _original_groups = _groups.copy()
             for group in list(_groups):
                 for parent_group in group.immediate_parent_groups:
-                    if cache_result:
-                        with self._dynamic_group_lock:
-                            self._dynamic_groups_resolved = None
-                            if (
-                                not parent_group.members_remove or
-                                not parent_group.members_remove(self)
-                            ):
-                                _groups.add(parent_group)
-                            self._dynamic_groups_resolved = True
-                    else:
-                        _groups.add(parent_group)
+                    _groups.add(parent_group)
             if _groups == _original_groups:
                 # we didn't add any new parent groups, so we can stop
                 break
 
-        if cache_result:
-            return sorted(_groups)
-        else:
-            raise DontCache(sorted(_groups))
-
-    @property
-    def _groups_dynamic(self):
-        """
-        Returns all groups whose members_add matches this node.
-        """
-        _groups = set()
-        for group in self.repo.groups:
-            if group.members_add is not None and group.members_add(self):
-                _groups.add(group)
-            if group.members_remove is not None and group.members_remove(self):
-                try:
-                    _groups.remove(group)
-                except KeyError:
-                    pass
         return _groups
 
     def has_any_bundle(self, bundle_list):
@@ -675,49 +677,57 @@ class Node(object):
 
     @property
     def metadata(self):
-        """
-        Returns full metadata for a node. MUST NOT be used from inside a
-        metadata processor. Use .partial_metadata instead.
-        """
-        if self._dynamic_groups_resolved is None:
-            # return only metadata set directly at the node level if
-            # we're still in the process of figuring out which groups
-            # we belong to
-            return self._node_metadata
-        else:
-            return self.repo._metadata_for_node(self.name, partial=False)
+        return self.repo._metadata_for_node(self.name)
 
     @property
     def metadata_blame(self):
-        return self.repo._metadata_for_node(self.name, partial=False, blame=True)
+        return self.repo._metadata_for_node(self.name, blame=True)
+
+    @property
+    def _metadata_stack(self):
+        return self.repo._metadata_for_node(self.name, stack=True)
+
+    def metadata_get(self, path, default=NO_DEFAULT):
+        if not isinstance(path, (tuple, list)):
+            path = path.split("/")
+        try:
+            return value_at_key_path(self.metadata, path)
+        except KeyError:
+            if default != NO_DEFAULT:
+                return default
+            else:
+                raise
 
     def metadata_hash(self):
         return hash_metadata(self.metadata)
 
     @property
-    def metadata_processors(self):
+    def metadata_defaults(self):
         for bundle in self.bundles:
-            for metadata_processor in bundle.metadata_processors:
+            if bundle._metadata_defaults_and_reactors[0]:
                 yield (
-                    "{}.{}".format(
+                    "metadata_defaults:{}".format(bundle.name),
+                    bundle._metadata_defaults_and_reactors[0],
+                )
+
+    @property
+    def metadata_reactors(self):
+        for bundle in self.bundles:
+            for reactor in bundle._metadata_defaults_and_reactors[1]:
+                yield (
+                    "metadata_reactor:{}.{}".format(
                         bundle.name,
-                        metadata_processor.__name__,
+                        reactor.__name__,
                     ),
-                    metadata_processor,
+                    reactor,
                 )
 
     @property
     def partial_metadata(self):
         """
-        Only to be used from inside metadata processors. Can't use the
-        normal .metadata there because it might deadlock when nodes
-        have interdependent metadata.
-
-        It's OK for metadata processors to work with partial metadata
-        because they will be fed all metadata updates until no more
-        changes are made by any metadata processor.
+        Deprecated, remove in 5.0.0
         """
-        return self.repo._metadata_for_node(self.name, partial=True)
+        return self.metadata
 
     def run(self, command, data_stdin=None, may_fail=False, log_output=False):
         assert self.os in self.OS_FAMILY_UNIX
@@ -763,19 +773,29 @@ class Node(object):
             wrapper_outer=self.cmd_wrapper_outer,
         )
 
-    @property
-    def template_node(self):
-        if not self._template_node_name:
-            return None
-        else:
-            target_node = self.repo.get_node(self._template_node_name)
-            if target_node._template_node_name:
-                raise RepositoryError(_(
-                    "{template_node} cannot use template_node because {node} uses {template_node} "
-                    "as template_node"
-                ).format(node=self.name, template_node=target_node.name))
-            else:
-                return target_node
+    @cached_property
+    def toml(self):
+        if not self.file_path or not self.file_path.endswith(".toml"):
+            raise ValueError(_("node {} not in TOML format").format(self.name))
+        return toml_parse(get_file_contents(self.file_path))
+
+    def toml_save(self):
+        try:
+            toml_doc = self.toml
+        except ValueError:
+            attributes = self._attributes.copy()
+            del attributes['file_path']
+            toml_doc = dict_to_toml(attributes)
+            self.file_path = join(self.repo.path, "nodes", self.name + ".toml")
+        if not exists(join(self.repo.path, "nodes")):
+            mkdir(join(self.repo.path, "nodes"))
+        with open(self.file_path, 'w') as f:
+            f.write(toml_clean(toml_dump(toml_doc)))
+
+    def toml_set(self, path, value):
+        if not isinstance(path, tuple):
+            path = path.split("/")
+        set_key_at_path(self.toml, path, value)
 
     def upload(self, local_path, remote_path, mode=None, owner="", group="", may_fail=False):
         assert self.os in self.OS_FAMILY_UNIX
@@ -823,10 +843,6 @@ def build_attr_property(attr, default):
                 attr_source = "group:{}".format(group.name)
                 attr_value = getattr(group, attr)
 
-        if self.template_node:
-            attr_source = "template_node"
-            attr_value = getattr(self.template_node, attr)
-
         if getattr(self, "_{}".format(attr)) is not None:
             attr_source = "node"
             attr_value = getattr(self, "_{}".format(attr))
@@ -840,12 +856,8 @@ def build_attr_property(attr, default):
             attr=attr,
             source=attr_source,
         ))
-        if self._dynamic_groups_resolved:
-            return attr_value
-        else:
-            raise DontCache(attr_value)
-    method.__name__ = str("_group_attr_{}".format(attr))  # required for cached_property
-                                                          # str() for Python 2 compatibility
+        return attr_value
+    method.__name__ = "_group_attr_{}".format(attr)  # required for cached_property
     return cached_property(method)
 
 for attr, default in GROUP_ATTR_DEFAULTS.items():
