@@ -1,6 +1,3 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from base64 import b64encode
 from codecs import getwriter
 from contextlib import contextmanager
@@ -8,6 +5,7 @@ import hashlib
 from inspect import isgenerator
 from os import chmod, close, makedirs, remove
 from os.path import dirname, exists
+from random import shuffle
 import stat
 from sys import stderr, stdout
 from tempfile import mkstemp
@@ -16,19 +14,11 @@ from requests import get
 
 from ..exceptions import DontCache, FaultUnavailable
 
-__GETATTR_CODE_CACHE = {}
-__GETATTR_RESULT_CACHE = {}
-__GETATTR_NODEFAULT = "very_unlikely_default_value"
-
 
+class NO_DEFAULT: pass
 MODE644 = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
-
-try:
-    STDERR_WRITER = getwriter('utf-8')(stderr.buffer)
-    STDOUT_WRITER = getwriter('utf-8')(stdout.buffer)
-except AttributeError:  # Python 2
-    STDERR_WRITER = getwriter('utf-8')(stderr)
-    STDOUT_WRITER = getwriter('utf-8')(stdout)
+STDERR_WRITER = getwriter('utf-8')(stderr.buffer)
+STDOUT_WRITER = getwriter('utf-8')(stdout.buffer)
 
 
 def cached_property(prop):
@@ -54,21 +44,60 @@ def cached_property(prop):
 
 
 def download(url, path):
-    if not exists(dirname(path)):
-        makedirs(dirname(path))
-    if exists(path):
-        chmod(path, MODE644)
-    with open(path, 'wb') as f:
-        r = get(url, stream=True)
-        r.raise_for_status()
-        for block in r.iter_content(1024):
-            if not block:
-                break
-            else:
-                f.write(block)
+    with error_context(url=url, path=path):
+        if not exists(dirname(path)):
+            makedirs(dirname(path))
+        if exists(path):
+            chmod(path, MODE644)
+        with open(path, 'wb') as f:
+            r = get(url, stream=True)
+            r.raise_for_status()
+            for block in r.iter_content(1024):
+                if not block:
+                    break
+                else:
+                    f.write(block)
 
 
-class Fault(object):
+class ErrorContext(Exception):
+    pass
+
+
+@contextmanager
+def error_context(**kwargs):
+    """
+    This can be used to provide context for critical exceptions. Since
+    we're processing lots of different dicts, a "KeyError: foo" will
+    often not be helpful, since it's not clear which dict is missing the
+    key.
+
+
+    >>> with error_context(arbitrary_kwarg="helpful hint"):
+    ...     {}["foo"]
+    ...
+    Traceback (most recent call last):
+      [...]
+    KeyError: 'foo'
+
+    The above exception was the direct cause of the following exception:
+
+    Traceback (most recent call last):
+      [...]
+    bundlewrap.utils.ErrorContext: ACTUAL EXCEPTION ABOVE
+    {'arbitrary_kwarg': 'helpful hint'}
+
+
+    Careful though: Only use this in places where you don't expect
+    exceptions to occur, since they will indiscriminately be reraised as
+    ErrorContext.
+    """
+    try:
+        yield
+    except Exception as exc:
+        raise ErrorContext("ACTUAL EXCEPTION ABOVE\n" + repr(kwargs)) from exc
+
+
+class Fault:
     """
     A proxy object for lazy access to things that may not really be
     available at the time of use.
@@ -76,7 +105,16 @@ class Fault(object):
     This let's us gracefully skip items that require information that's
     currently not available.
     """
-    def __init__(self, callback, **kwargs):
+    def __init__(self, fault_identifier, callback, **kwargs):
+        if isinstance(fault_identifier, list):
+            self.id_list = fault_identifier
+        else:
+            self.id_list = [fault_identifier]
+
+        for key, value in sorted(kwargs.items()):
+            self.id_list.append(hash(key))
+            self.id_list.append(hash(value))
+
         self._available = None
         self._exc = None
         self._value = None
@@ -98,11 +136,20 @@ class Fault(object):
         if isinstance(other, Fault):
             def callback():
                 return self.value + other.value
-            return Fault(callback)
+            return Fault(self.id_list + other.id_list, callback)
         else:
             def callback():
                 return self.value + other
-            return Fault(callback)
+            return Fault(self.id_list + ['raw {}'.format(repr(other))], callback)
+
+    def __eq__(self, other):
+        if not isinstance(other, Fault):
+            return False
+        else:
+            return self.id_list == other.id_list
+
+    def __hash__(self):
+        return hash(tuple(self.id_list))
 
     def __len__(self):
         return len(self.value)
@@ -116,12 +163,12 @@ class Fault(object):
     def b64encode(self):
         def callback():
             return b64encode(self.value.encode('UTF-8')).decode('UTF-8')
-        return Fault(callback)
+        return Fault(self.id_list + ['b64encode'], callback)
 
     def format_into(self, format_string):
         def callback():
             return format_string.format(self.value)
-        return Fault(callback)
+        return Fault(self.id_list + ['format_into ' + format_string], callback)
 
     @property
     def is_available(self):
@@ -140,7 +187,7 @@ def _make_method_callback(method_name):
     def method(self, *args, **kwargs):
         def callback():
             return getattr(self.value, method_name)(*args, **kwargs)
-        return Fault(callback)
+        return Fault(self.id_list + [method_name], callback)
     return method
 
 
@@ -158,55 +205,12 @@ for method_name in (
 
 
 def get_file_contents(path):
-    with open(path, 'rb') as f:
-        content = f.read()
+    with error_context(path=path):
+        with open(path, 'rb') as f:
+            content = f.read()
     return content
 
 
-def get_all_attrs_from_file(path, base_env=None):
-    """
-    Reads all 'attributes' (if it were a module) from a source file.
-    """
-    if base_env is None:
-        base_env = {}
-
-    if not base_env and path in __GETATTR_RESULT_CACHE:
-        # do not allow caching when passing in a base env because that
-        # breaks repeated calls with different base envs for the same
-        # file
-        return __GETATTR_RESULT_CACHE[path]
-
-    if path not in __GETATTR_CODE_CACHE:
-        source = get_file_contents(path)
-        __GETATTR_CODE_CACHE[path] = compile(source, path, mode='exec')
-
-    code = __GETATTR_CODE_CACHE[path]
-    env = base_env.copy()
-    try:
-        exec(code, env)
-    except:
-        from .ui import io
-        io.stderr("Exception while executing {}".format(path))
-        raise
-
-    if not base_env:
-        __GETATTR_RESULT_CACHE[path] = env
-
-    return env
-
-
-def getattr_from_file(path, attrname, base_env=None, default=__GETATTR_NODEFAULT):
-    """
-    Reads a specific 'attribute' (if it were a module) from a source
-    file.
-    """
-    env = get_all_attrs_from_file(path, base_env=base_env)
-    if default == __GETATTR_NODEFAULT:
-        return env[attrname]
-    else:
-        return env.get(attrname, default)
-
-
 def hash_local_file(path):
     """
     Retuns the sha1 hash of a file on the local machine.
@@ -225,6 +229,15 @@ def names(obj_list):
         yield obj.name
 
 
+def randomize_order(obj):
+    if isinstance(obj, dict):
+        result = list(obj.items())
+    else:
+        result = list(obj)
+    shuffle(result)
+    return result
+
+
 def sha1(data):
     """
     Returns hex SHA1 hash for input.
@@ -234,7 +247,7 @@ def sha1(data):
     return hasher.hexdigest()
 
 
-class SkipList(object):
+class SkipList:
     """
     Used to maintain a list of nodes that have already been visited.
     """
