@@ -1,7 +1,4 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
-from pipes import quote
+from shlex import quote
 
 from . import cached_property
 from .text import force_text, mark_for_translation as _
@@ -39,7 +36,7 @@ def stat(node, path):
     return file_stat
 
 
-class PathInfo(object):
+class PathInfo:
     """
     Serves as a proxy to get_path_type.
     """
@@ -60,10 +57,6 @@ class PathInfo(object):
     def group(self):
         return self.stat['group']
 
-    @property
-    def is_binary_file(self):
-        return self.is_file and not self.is_text_file
-
     @property
     def is_directory(self):
         return self.stat['type'] == "directory"
@@ -114,16 +107,6 @@ class PathInfo(object):
         # contains backslash-escaped characters – we must lstrip() that
         return force_text(result.stdout).strip().lstrip("\\").split()[0]
 
-    @cached_property
-    def sha256(self):
-        if self.node.os == 'macos':
-            result = self.node.run("shasum -a 256 -- {}".format(quote(self.path)))
-        elif self.node.os in self.node.OS_FAMILY_BSD:
-            result = self.node.run("sha256 -q -- {}".format(quote(self.path)))
-        else:
-            result = self.node.run("sha256sum -- {}".format(quote(self.path)))
-        return force_text(result.stdout).strip().split()[0]
-
     @property
     def size(self):
         return self.stat['size']
