@@ -1,16 +1,11 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from contextlib import contextmanager
 from datetime import datetime
-from errno import EPIPE
-import fcntl
 from functools import wraps
 from os import _exit, environ, getpid, kill
 from os.path import join
 from select import select
+from shutil import get_terminal_size
 from signal import signal, SIG_DFL, SIGINT, SIGQUIT, SIGTERM
-import struct
 from subprocess import PIPE, Popen
 import sys
 import termios
@@ -37,12 +32,6 @@ SHUTDOWN_EVENT_SOFT = Event()
 TTY = STDOUT_WRITER.isatty()
 
 
-if sys.version_info >= (3, 0):
-    broken_pipe_exception = BrokenPipeError
-else:
-    broken_pipe_exception = IOError
-
-
 def add_debug_indicator(f):
     @wraps(f)
     def wrapped(self, msg, **kwargs):
@@ -112,26 +101,27 @@ def spinner():
             yield c
 
 
-def term_width():
-    if not TTY:
-        return 0
-
-    fd = sys.stdout.fileno()
-    _, width = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, 'aaaa'))
-    return width
-
-
 def page_lines(lines):
     """
     View the given list of Unicode lines in a pager (e.g. `less`).
     """
     lines = list(lines)
-    line_width = max([len(ansi_clean(line)) for line in lines])
-    if TTY and line_width > term_width():
-        pager = Popen([environ.get("PAGER", "/usr/bin/less")], stdin=PIPE)
-        pager.stdin.write("\n".join(lines).encode('utf-8'))
+    if TTY:
+        write_to_stream(STDOUT_WRITER, SHOW_CURSOR)
+        env = environ.copy()
+        env["LESS"] = env.get("LESS", "") + " -FR"
+        pager = Popen(
+            [environ.get("PAGER", "/usr/bin/less")],
+            env=env,
+            stdin=PIPE,
+        )
+        try:
+            pager.stdin.write("\n".join(lines).encode('utf-8'))
+        except BrokenPipeError:
+            pass
         pager.stdin.close()
         pager.communicate()
+        write_to_stream(STDOUT_WRITER, HIDE_CURSOR)
     else:
         for line in lines:
             io.stdout(line)
@@ -144,13 +134,11 @@ def write_to_stream(stream, msg):
         else:
             stream.write(ansi_clean(msg))
         stream.flush()
-    except broken_pipe_exception as e:
-        if broken_pipe_exception == IOError:
-            if e.errno != EPIPE:
-                raise
+    except BrokenPipeError:
+        pass
 
 
-class DrainableStdin(object):
+class DrainableStdin:
     def get_input(self):
         while True:
             if QUIT_EVENT.is_set():
@@ -163,7 +151,7 @@ class DrainableStdin(object):
             termios.tcflush(sys.stdin, termios.TCIFLUSH)
 
 
-class IOManager(object):
+class IOManager:
     """
     Threadsafe singleton class that handles all IO.
     """
@@ -428,7 +416,7 @@ class IOManager(object):
                 progress_text = "{:.1f}%  ".format(progress * 100)
                 line += bold(progress_text)
                 visible_length += len(progress_text)
-            line += self.jobs[-1][:term_width() - 1 - visible_length]
+            line += self.jobs[-1][:get_terminal_size().columns - 1 - visible_length]
             write_to_stream(STDOUT_WRITER, line)
             self._status_line_present = True
 
